@@ -1,3 +1,3 @@
 """Microsoft Word review tools (comments, markup, etc.) with Python."""
 
-from .docxrev import *  # noqa: F401, F403
+from docxrev.docxrev import *  # noqa: F401, F403
